import torch

from metlibvi.plotting import plot_true_prop_accept


def on_transition_end_callback(z_target=None, z_accepted=None, z_proposal=None, acceptance_logprobs=None, verbosity=1,
                               current_iter=1, use_callback=False, **kwargs):
    if use_callback:
        if current_iter % verbosity == 0:
            with torch.no_grad():
                plot_true_prop_accept(z_target=z_target, z_accepted=z_accepted, z_proposal=z_proposal,
                                      acceptance_logprobs=acceptance_logprobs,
                                      **kwargs)
                acceptance_rate = ((z_accepted == z_proposal) * 1.).mean(-1).mean()
                print('Mean acceptance rate is', acceptance_rate)



import torch


def get_acceptance_ratio(log_t, log_1pt, use_barker):
    """
    The function computes acceptance ratio, amenable for backpropagation
    :param log_t: logarithm of acceptance ratio
    :param log_1pt: logarithm of 1 + t, computed in numerically accurate way
    :param use_barker: whether to use barker ratio or not
    :return: decision, corresponding log_ratios and acceptance log ratios
    """
    if use_barker:
        current_log_alphas_pre = log_t - log_1pt
    else:
        current_log_alphas_pre = torch.min(log_t, torch.zeros_like(log_t))

    log_probs = torch.log(torch.rand_like(log_t))
    a = log_probs <= current_log_alphas_pre
    if use_barker:
        current_log_alphas = current_log_alphas_pre.clone()
        current_log_alphas[~a] = (-log_1pt)[~a]
    else:
        expression = torch.ones_like(current_log_alphas_pre) - torch.exp(current_log_alphas_pre)
        corr_expression = torch.log(expression + 1e-8)
        current_log_alphas = current_log_alphas_pre.clone()
        current_log_alphas[~a] = corr_expression[~a]

    return a, current_log_alphas, current_log_alphas_pre.detach()


def compute_grad(z, target_logprob, x):
    flag = z.requires_grad  # True, if requires grad (means that we propagate gradients to some parameters)
    if not flag:
        z_ = z.detach().requires_grad_(True)
    else:
        z_ = z
    with torch.enable_grad():
        grad = get_grad(z=z_, target_logprob=target_logprob, x=x)
        if not flag:
            grad = grad.detach()
            z_.requires_grad_(False)
        return grad


def get_grad(z, target_logprob, x=None):
    s = target_logprob(x=x, z=z)
    grad = torch.autograd.grad(s.sum(), z, create_graph=True, only_inputs=True)[0]
    return grad


def run_chain(kernel, z_init, target_logprob, x=None, n_steps=100, return_trace=False, burnin=0):
    """
    The function launches mcmc chain.
    :param kernel: mcmc kernel
    :param z_init: initial state
    :param target_logprob: target_logprob (must have log_prob method, which returns a batch)
    :param x: data object (if given)
    :param n_steps: number of transitions
    :param return_trace: whether to return trajectory (True), or only final state (False)
    :param burnin: number of burnin steps
    :return: either trajectory or final samples
    """
    samples = z_init
    if not return_trace:
        for _ in range(burnin + n_steps):
            samples = kernel(z=samples, target_logprob=target_logprob, x=x)["z_new"].detach()
        return samples
    else:
        final = torch.tensor([], device=z_init.device, dtype=torch.float32)
        for i in range(burnin + n_steps):
            samples = kernel(z=samples, target_logprob=target_logprob, x=x)["z_new"].detach()
            if i >= burnin:
                final = torch.cat([final, samples])
        return final



import numpy as np
import torch
import torch.nn as nn
from tqdm.auto import tqdm

from metlibvi.train_utils import metropolized_vi_forward, metropolized_loss, vanilla_vi_forward, kl_loss


def train_met_vi(transitions, target, prior, n_batches=1000, batch_size=150, lr=5e-3):
    if transitions[0].name == "RealNVP":
        logit_p = nn.Parameter(torch.tensor(np.zeros(len(transitions)), dtype=torch.float32))
        additional_params = [logit_p]
    else:
        logit_p = None
        additional_params = []
    optimizer = torch.optim.Adam(list(transitions.parameters()) + additional_params, lr=lr)
    for _ in tqdm(range(n_batches)):
        z = prior.sample((batch_size,))
        output = metropolized_vi_forward(transitions=transitions, forward_logit_prob=logit_p,
                                         target_logprob=target.log_prob, z=z)
        loss = metropolized_loss(final_samples=output["z_new"], z=z, x=None, prior=prior,
                                 sum_log_jac=output["aggregated_log_jac"],
                                 sum_log_alphas=output["sum_log_alphas"], sum_log_probs=output["sum_log_probs"],
                                 target_logprob=target.log_prob,
                                 loss_variant="standard")
        loss.backward()
        optimizer.step()
        optimizer.zero_grad()

    return transitions, logit_p


def train_vi(transitions, target, prior, n_batches=1000, batch_size=150, lr=5e-3):
    optimizer = torch.optim.Adam(list(transitions.parameters()), lr=lr)
    for _ in tqdm(range(n_batches)):
        z = prior.sample((batch_size,))
        output = vanilla_vi_forward(transitions=transitions, z=z, )
        kl = kl_loss(final_samples=output["z_new"], z=z, prior=prior, sum_log_jac=output["aggregated_log_jac"],
                     target_logprob=target.log_prob, loss_variant="standard", x=None)
        kl.backward()
        optimizer.step()
        optimizer.zero_grad()

    return transitions



import pytest
import torch

from metlibvi.targets import Gaussian

device = torch.device("cpu")  # "cuda" if torch.cuda.is_available() else "cpu")


@pytest.fixture(scope="session")
def get_gaussian_target():
    loc = torch.tensor([10., -10.], device=device)
    cov = torch.eye(2, device=device)
    target = Gaussian(loc=loc, covariance_matrix=cov)
    return target



import torch


def metropolized_loss(final_samples, z, x, prior, sum_log_jac, sum_log_alphas, sum_log_probs, target_logprob,
                      loss_variant):
    if loss_variant == 'standard':
        elbo = target_logprob(z=final_samples, x=x) - (prior.log_prob(z) - sum_log_jac + sum_log_alphas)
        grad_elbo = torch.mean(elbo + elbo.detach() * (sum_log_alphas + sum_log_probs))
        return -grad_elbo


def kl_loss(final_samples, z, x, prior, sum_log_jac, target_logprob, loss_variant):
    if loss_variant == 'standard':
        return torch.mean(prior.log_prob(z) - sum_log_jac - target_logprob(z=final_samples, x=x))



import torch
import torch.nn as nn
from pyro.distributions.transforms import AffineAutoregressive, BlockAutoregressive, AffineCoupling, \
    ConditionalAffineCoupling
from pyro.nn import AutoRegressiveNN, DenseNN, ConditionalDenseNN


class NormFlow(nn.Module):
    def __init__(self, flow_type, num_flows, hidden_dim=20, need_permute=False, **kwargs):
        super(NormFlow, self).__init__()
        self.need_permute = need_permute
        self.name = flow_type
        self.condition = False
        if flow_type == 'IAF':
            self.flow = nn.ModuleList(
                [AffineAutoregressive(AutoRegressiveNN(input_dim=hidden_dim, param_dims=[2 * hidden_dim], **kwargs),
                                      stable=True) for _ in
                 range(num_flows)])
        elif flow_type == 'BNAF':
            self.flow = nn.ModuleList(
                [BlockAutoregressive(input_dim=hidden_dim, **kwargs) for _ in
                 range(num_flows)])
        elif flow_type == 'RealNVP':
            split_dim = hidden_dim // 2
            param_dims = [hidden_dim - split_dim, hidden_dim - split_dim]
            self.flow = nn.ModuleList(
                [AffineCoupling(split_dim, DenseNN(input_dim=split_dim, param_dims=param_dims, **kwargs)) for _ in
                 range(num_flows)])
        elif flow_type == "CondRealNVP":
            split_dim = hidden_dim // 2
            param_dims = [hidden_dim - split_dim, hidden_dim - split_dim]
            self.flow = nn.ModuleList(
                [ConditionalAffineCoupling(split_dim,
                                           ConditionalDenseNN(input_dim=split_dim, param_dims=param_dims, **kwargs)) for
                 _ in
                 range(num_flows)])
            self.condition = True
        else:
            raise NotImplementedError
        even = [i for i in range(0, hidden_dim, 2)]
        odd = [i for i in range(1, hidden_dim, 2)]
        undo_eo = [i // 2 if i % 2 == 0 else (i // 2 + len(even)) for i in range(hidden_dim)]
        undo_oe = [(i // 2 + len(odd)) if i % 2 == 0 else i // 2 for i in range(hidden_dim)]
        self.register_buffer('eo', torch.tensor(even + odd, dtype=torch.int64))
        self.register_buffer('oe', torch.tensor(odd + even, dtype=torch.int64))
        self.register_buffer('undo_eo', torch.tensor(undo_eo, dtype=torch.int64))
        self.register_buffer('undo_oe', torch.tensor(undo_oe, dtype=torch.int64))

    def permute(self, z, i, undo=False):
        if not undo:
            if i % 2 == 0:
                z = torch.index_select(z, 1, self.eo)
            else:
                z = torch.index_select(z, 1, self.oe)
        else:
            if i % 2 == 0:
                z = torch.index_select(z, 1, self.undo_eo)
            else:
                z = torch.index_select(z, 1, self.undo_oe)
        return z

    def forward(self, z, **kwargs):
        """
        Forward method
        :param z: initial state
        :return: a pair -- z_new and log_jacobian of the resulting transformation
        """
        log_jacob = torch.zeros_like(z[:, 0], dtype=torch.float32)
        for i in range(len(self.flow)):
            if self.need_permute:
                z = self.permute(z, i)

            if not self.condition:
                current_flow = self.flow[i]
            else:
                current_flow = self.flow[i].condition(**kwargs)

            z_new = current_flow(z)
            log_jacob += current_flow.log_abs_det_jacobian(z, z_new)

            if self.need_permute:
                z_new = self.permute(z_new, i, undo=True)
            z = z_new
        return {"z_new": z, "aggregated_log_jac": log_jacob}

    def inverse(self, z, **kwargs):
        """
        Inverse method (effective only for affine transform)
        :param z: initial state
        :return: a pair -- z_new and log_jacobian of the resulting transformation
        """
        log_jacob = torch.zeros_like(z[:, 0], dtype=torch.float32)
        for i in range(len(self.flow))[::-1]:
            if self.need_permute:
                z = self.permute(z, i)

            if not self.condition:
                current_flow = self.flow[i]
            else:
                current_flow = self.flow[i].condition(**kwargs)

            z_new = current_flow._inverse(z)
            log_jacob -= current_flow.log_abs_det_jacobian(z, z_new)
            if self.need_permute:
                z_new = self.permute(z_new, i, undo=True)
            z = z_new
        return {"z_new": z, "aggregated_log_jac": log_jacob}



import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
import torch


class BaseTarget:
    """
    Class for base target
    """

    def log_prob(self, z, x=None):
        """
        The function returns logarithmic density of point z given x (if any)
        :param z: Input state
        :param x: Input object
        :return: logarithmic density. Tensor of batch size (z.shape[0])
        """
        pass

    def sample(self, shape):
        """

        :param shape: The function returns samples of size shape (iterable).
        :return:
        """
        pass

    def plot_distr(self, lims, npts, device, figsize=(10, 4), dpi=100):
        x_lim = lims[0]
        y_lim = lims[1]

        xside = np.linspace(-x_lim - 1, x_lim + 1, npts)
        yside = np.linspace(-y_lim - 1, y_lim + 1, npts)
        xx, yy = np.meshgrid(xside, yside)

        z = torch.tensor(np.hstack([xx.reshape(-1, 1), yy.reshape(-1, 1)]), device=device, dtype=torch.float32)
        logdens = self.log_prob(z)
        p = np.exp(logdens.cpu().detach().numpy().reshape(npts, npts))

        fig, ax = plt.subplots(ncols=2, figsize=figsize, sharex='all', sharey='all', dpi=dpi)
        ax[0].set_title('Target density')
        ax[0].pcolormesh(xx, yy, p, norm=mcolors.PowerNorm(0.5))
        ax[0].contour(xx, yy, p, 15, colors='black', alpha=0.2)
        ax[0].set_xlim(-x_lim, x_lim)
        ax[0].set_ylim(0, y_lim)
        ax[0].set_xticks(np.arange(-x_lim, x_lim + 1, 10))
        ax[0].set_yticks(np.arange(-y_lim, y_lim + 1, 10))

        ax[1].set_title('Target samples')
        samples = self.sample((5000,)).cpu().detach().numpy()
        ax[1].scatter(samples[:, 0], samples[:, 1], alpha=0.075)
        ax[1].set_xticks(np.arange(-x_lim, x_lim + 1, 10))
        ax[1].set_yticks(np.arange(-y_lim, y_lim + 1, 10))


class Gaussian(BaseTarget):
    def __init__(self, loc, covariance_matrix):
        super().__init__()
        self.loc = loc
        self.covariance_matrix = covariance_matrix
        self.distribution = torch.distributions.MultivariateNormal(loc=self.loc,
                                                                   covariance_matrix=self.covariance_matrix)

    def log_prob(self, z, x=None):
        return self.distribution.log_prob(z)

    def sample(self, shape):
        return self.distribution.sample(shape)


class GMM(BaseTarget):
    def __init__(self, locs, covariance_matrices):
        super().__init__()
        self.locs = torch.tensor([], device=locs[0].device)
        self.covariance_matrices = torch.tensor([], device=covariance_matrices[0].device)
        self.distributions = []
        self.device = locs[0].device
        for i in range(len(locs)):
            self.distributions.append(
                torch.distributions.MultivariateNormal(loc=locs[i], covariance_matrix=covariance_matrices[i]))

    def log_prob(self, z, x=None):
        log_p = torch.tensor([], device=z.device)
        for i in range(len(self.distributions)):
            log_paux = self.distributions[i].log_prob(z).view(-1, 1)
            log_p = torch.cat([log_p, log_paux], dim=-1)
        log_density = torch.logsumexp(log_p, dim=1)
        return log_density

    def sample(self, shape):
        p = np.random.choice(a=len(self.distributions), size=shape[0])
        samples = torch.tensor([], device=self.device)
        for idx in p:
            z = self.distributions[idx].sample((1,))
            samples = torch.cat([samples, z])
        return samples


class Funnel(BaseTarget):
    def __init__(self, dim, device):
        super().__init__()
        self.d = dim
        self.device = device
        self.std_normal = torch.distributions.Normal(loc=torch.tensor(0., dtype=torch.float32, device=self.device),
                                                     scale=torch.tensor(1., dtype=torch.float32, device=self.device), )

    def log_prob(self, z, x=None):
        d = z.shape[1]
        fst_component = z[:, 0]
        log_density = -fst_component ** 2 / 2 - torch.sum(z[:, 1:] ** 2., dim=1) * torch.exp(
            -2. * fst_component) / 2. - (d - 1) * fst_component
        return log_density

    def sample(self, shape):
        samples = torch.zeros((shape[0], self.d), device=self.device)
        for i in range(shape[0]):
            samples[i][0] = self.std_normal.sample((1,))
            component_normal = torch.distributions.Normal(
                loc=torch.zeros(self.d - 1, device=self.device, dtype=torch.float32),
                scale=torch.exp(samples[i][0]) * torch.ones(self.d - 1, device=self.device, dtype=torch.float32))
            samples[i][1:] = component_normal.sample()
        return samples


class Banana(BaseTarget):
    def __init__(self, dim, a, b, rho, device):
        super().__init__()
        self.device = device
        self.initial_gaussian = torch.distributions.MultivariateNormal(loc=torch.zeros(dim, device=self.device),
                                                                       covariance_matrix=rho)
        self.a = a
        self.b = b
        self.rho = rho[0, 1]

    def log_prob(self, z, x=None):
        x = z[:, 0]
        y = z[:, 1]
        log_density = -1. / (2 * (1. - self.rho ** 2)) * ((x / self.a) ** 2
                                                          + self.a ** 2 * (
                                                                  y - self.b * x ** 2 / self.a ** 2 - self.b * self.a ** 2) ** 2
                                                          - 2 * self.rho * (
                                                                  y - self.b * x ** 2 / self.a ** 2 - self.b * self.a ** 2))
        return log_density

    def sample(self, shape):
        # sample from true target
        gaussian_samples = self.initial_gaussian.sample(shape)
        x = self.a * gaussian_samples[:, 0]
        y = gaussian_samples[:, 1] / self.a + self.b * (gaussian_samples[:, 0] ** 2 + self.a ** 2)
        samples = torch.cat([x[:, None], y[:, None]], dim=-1)
        return samples


class BananaMixture(BaseTarget):
    # TODO: Debug this
    def __init__(self, dim, locs, cov, a, b, rho, device):
        super().__init__()
        self.device = device
        self.locs = locs
        self.device = device
        self.a = a
        self.b = b
        self.logprob = torch.tensor(np.log(1. / len(self.locs)), dtype=torch.float32, device=self.device)
        self.rho = rho[0, 1]
        self.initial_gaussian = torch.distributions.MultivariateNormal(loc=torch.zeros(dim, device=self.device),
                                                                       covariance_matrix=cov)

    def log_prob(self, z, x=None):
        x = z[:, 0]
        y = z[:, 1]
        log_density = torch.tensor([], device=self.device)
        for i in range(len(self.locs)):
            x_c = x - self.locs[i][0]
            y_c = y - self.locs[i][1]
            log_paux = self.logprob - 1. / (2 * (1. - self.rho ** 2)) * ((x_c / self.a[i]) ** 2
                                                                         + self.a[i] ** 2 * (
                                                                                 y_c - self.b[i] * x_c ** 2 / self.a[
                                                                             i] ** 2 - self.b[
                                                                                     i] * self.a[i] ** 2) ** 2
                                                                         - 2 * self.rho * (
                                                                                 y_c - self.b[i] * x_c ** 2 / self.a[
                                                                             i] ** 2 - self.b[i] *
                                                                                 self.a[i] ** 2))

            log_density = torch.cat([log_density, log_paux[None, :]], dim=0)
        log_density = torch.logsumexp(log_density, dim=0)
        return log_density

    def sample(self, shape):
        # sample from true target
        gaussian_samples = self.initial_gaussian.sample(shape)
        p = np.random.choice(a=len(self.locs), size=shape[0])
        x = self.a[p] * (gaussian_samples[:, 0] + self.locs[p, 0])
        y = gaussian_samples[:, 1] / self.a[p] + self.b[p] * (
                gaussian_samples[:, 0] ** 2 + self.a[p] ** 2)  ##Only valid for mu_y = 0 !!
        samples = torch.cat([x[:, None], y[:, None]], dim=-1)
        return samples



import torch

from metlibvi.callbacks import on_transition_end_callback
from metlibvi.mcmc import get_acceptance_ratio


def vanilla_vi_forward(transitions, z, x=None, **kwargs):
    """
    The function implements forward step for vanilla variational inference
    :param transitions: transitions, represented ad a Modulelist
    :param z: a batch of initial (not yet transformed) samples (e.g. from standard normal)
    :param x: a batch of data objects (if any)
    :return: returns mean KL divergence
    """

    output = transitions(z.clone())
    final_samples, sum_log_jac = output["z_new"], output["aggregated_log_jac"]

    return {"z_new": final_samples, "aggregated_log_jac": sum_log_jac}


def transition_forward_aux(transition, input_dict):
    if transition.name == "RealNVP":
        with torch.no_grad():
            prob = torch.sigmoid(input_dict["forward_logit_prob"])
            rand = torch.rand_like(prob)
            move_forward = rand < prob.squeeze()
        f = transition(**input_dict)
        b = transition.inverse(**input_dict)

        f["forward_logit_prob"] = input_dict["forward_logit_prob"]
        b["forward_logit_prob"] = -input_dict["forward_logit_prob"]

        output = apply_acceptance(current_state=b, proposal_state=f, decisions=move_forward)
        # this is to simplify log acceptance computation. There should be + or - forward logit
        multiplier = move_forward * 2. - 1.
    else:
        output = transition(**input_dict)
        output["forward_logit_prob"] = input_dict["forward_logit_prob"]
        multiplier = torch.ones_like(output["z_new"][:, 0])
    return output, multiplier


def apply_decision_unit(current, proposed, decisions):
    res = torch.empty_like(current)
    res[decisions] = proposed[decisions]
    res[~decisions] = current[~decisions]
    return res


def apply_acceptance(current_state, proposal_state, decisions):
    result = {}
    for k in current_state.keys():
        result[k] = apply_decision_unit(current_state[k], proposal_state[k], decisions)

    return result


def metropolized_vi_forward(transitions, target_logprob, z, x=None, forward_logit_prob=None,
                            accept_reject_correction=True,
                            use_barker=False, **kwargs):
    """
    The function returns forward step for metropolized vi
    :param transitions: transitions, represented ad a Modulelist
    :param target_logprob: a function, which computes target's logprob
    :param forward_logit_prob: a vector of length len(transitions), which contains logit probabilities of forward transitions
    :param z: a batch of initial (not yet transformed) samples (e.g. from standard normal)
    :param x: a batch of data objects (if any)
    :param use_barker: whether to use barker ratio or not
    :return: returns loss (precomputed grad) for metropolized algorithms
    """
    if forward_logit_prob is None:
        forward_logit_prob = torch.zeros(len(transitions))

    sum_log_alphas = torch.zeros_like(z[:, 0])
    sum_log_jac = torch.zeros_like(z[:, 0])
    sum_log_probs = torch.zeros_like(z[:, 0])

    z_cur = z.clone()

    for i in range(len(transitions)):
        input_dict = {"z": z_cur, "target_logprob": target_logprob, "x": x,
                      "forward_logit_prob": forward_logit_prob[i] * torch.ones_like(z[:, 0])}
        output, multiplier = transition_forward_aux(transition=transitions[i], input_dict=input_dict)

        target_log_density_upd = target_logprob(z=output["z_new"], x=x)
        target_log_density_old = target_logprob(z=z, x=x)

        log_t = target_log_density_upd - target_log_density_old - input_dict["forward_logit_prob"] * multiplier + \
                output["aggregated_log_jac"] + output.get("aux_for_acceptance_ratio", torch.zeros_like(z[:, 0]))
        log_1pt = torch.logsumexp(torch.cat([torch.zeros_like(log_t).view(-1, 1),
                                             log_t.view(-1, 1)], dim=-1), dim=-1)

        a, current_log_alphas, acceptance_log_alphas = get_acceptance_ratio(log_t, log_1pt, use_barker=use_barker)

        if accept_reject_correction:
            current_state = {"z_new": z_cur, "aggregated_log_jac": torch.zeros_like(z[:, 0]),
                             "forward_logit_prob": input_dict["forward_logit_prob"]}
            resulting_dict = apply_acceptance(current_state=current_state, proposal_state=output, decisions=a)
        else:
            resulting_dict = output

        sum_log_alphas = sum_log_alphas + current_log_alphas
        sum_log_jac = sum_log_jac + resulting_dict["aggregated_log_jac"]
        sum_log_probs = sum_log_probs + torch.log(torch.sigmoid(resulting_dict["forward_logit_prob"]))

        on_transition_end_callback(z_accepted=resulting_dict["z_new"], z_proposal=output["z_new"],
                                   acceptance_logprobs=current_log_alphas,
                                   **kwargs)

        z_cur = resulting_dict["z_new"]

    return {"z_new": z_cur, "aggregated_log_jac": sum_log_jac, "sum_log_probs": sum_log_probs,
            "sum_log_alphas": sum_log_alphas, }



import torch
import torch.nn as nn

from metlibvi.mcmc.utils import get_acceptance_ratio, compute_grad


class BaseKernel(nn.Module):
    """
    Base mcmc kernel class
    """

    def __init__(self, step_size, use_barker=False, learnable=False):
        """

        :param step_size: stepsize for transition
        :param use_barker: If True -- Barker ratios applied. MH otherwise
        :param learnable: whether learnable (usage for Met model) or not
        """
        super().__init__()
        self.use_barker = use_barker
        self.learnable = learnable
        self.register_buffer('zero', torch.tensor(0., dtype=torch.float32))
        self.register_buffer('one', torch.tensor(1., dtype=torch.float32))
        self.log_stepsize = nn.Parameter(torch.log(torch.tensor(step_size, dtype=torch.float32)),
                                         requires_grad=learnable)

    @property
    def step_size(self):
        return torch.exp(self.log_stepsize)


class HMC(BaseKernel):
    def __init__(self, n_leapfrogs, partial_ref=False, **kwargs):
        """
        HMC kernel class
        :param n_leapfrogs: number of leapfrog iterations
        :param partial_ref: whether use partial refresh or not
        """
        super().__init__(**kwargs)
        self.n_leapfrogs = n_leapfrogs
        self.partial_ref = partial_ref
        self.alpha_logit = nn.Parameter(self.zero, requires_grad=self.learnable)  # for partial refreshment
        self.name = "HMC"

    @property
    def alpha(self):
        return torch.sigmoid(self.alpha_logit)

    def _forward_step(self, z_old, x=None, target_logprob=None, p_old=None):
        p_ = p_old + self.step_size / 2. * compute_grad(z=z_old, target_logprob=target_logprob,
                                                        x=x)
        z_ = z_old
        for current_l in range(self.n_leapfrogs):
            z_ = z_ + self.step_size * p_
            if current_l != self.n_leapfrogs - 1:
                p_ = p_ + self.step_size * compute_grad(z=z_, target_logprob=target_logprob,
                                                        x=x)
        p_ = p_ + self.step_size / 2. * compute_grad(z=z_, target_logprob=target_logprob,
                                                     x=x)
        return z_, p_

    def _make_transition(self, z_old, target_logprob, p_old=None, x=None):

        std_normal = torch.distributions.Normal(loc=self.zero, scale=self.one)

        ############ Then we compute new points and densities ############
        z_upd, p_upd = self.forward_step(z_old=z_old, p_old=p_old, target_logprob=target_logprob, x=x)

        target_log_density_f = target_logprob(z=z_upd, x=x) + std_normal.log_prob(p_upd).sum(-1)
        target_log_density_old = target_logprob(z=z_old, x=x) + std_normal.log_prob(p_old).sum(-1)

        log_t = target_log_density_f - target_log_density_old
        log_1pt = torch.logsumexp(torch.cat([torch.zeros_like(log_t).view(-1, 1),
                                             log_t.view(-1, 1)], dim=-1), dim=-1)  # log(1+t)

        a, current_log_alphas, _ = get_acceptance_ratio(log_t=log_t, log_1pt=log_1pt, use_barker=self.use_barker)

        z_new = z_upd
        z_new[~a] = z_old[~a]

        p_new = -p_upd
        p_new[~a] = -p_old[~a]

        return z_new, p_new, a.to(torch.float32), current_log_alphas

    def forward_step(self, z_old, x=None, target_logprob=None, p_old=None):
        z_, p_ = self._forward_step(z_old=z_old, x=x, target_logprob=target_logprob, p_old=p_old)
        return z_, p_

    def forward(self, z, target_logprob, x=None, p=None):
        """

        :param z: current state
        :param target_logprob: target log prob
        :param x: batch of objects
        :param p: current momentum
        :return: (new state, new momentum, decisions casted to float, logs of corresponding alphas
        """
        if p is None:
            p = torch.randn_like(z)
        if self.partial_ref:
            p = p * self.alpha + torch.sqrt(self.one - self.alpha ** 2) * torch.randn_like(p)
        z_new, p_new, a, current_log_alphas = self._make_transition(z_old=z,
                                                                    target_logprob=target_logprob, p_old=p, x=x)
        return {"z_new": z_new,
                "p_new": p_new,
                "decisions": a.to(torch.float32),
                "current_log_alphas": current_log_alphas}


class MALA(BaseKernel):
    def __init__(self, **kwargs):
        """
        MALA kernel class

        :param use_barker: If True -- Barker ratios applied. MH otherwise
        """
        super().__init__(**kwargs)
        self.name = "MALA"

    def _forward_step(self, z_old, x=None, target_logprob=None):
        eps = torch.randn_like(z_old)
        forward_grad = compute_grad(z=z_old,
                                    target_logprob=target_logprob,
                                    x=x)
        update = torch.sqrt(2 * self.step_size) * eps + self.step_size * forward_grad
        return z_old + update, update, eps, forward_grad

    def forward(self, z, target_logprob, x=None):
        """

        :param z: current state
        :param target_logprob: target log prob
        :param x: batch of objects
        :return: (new state, decisions casted to float, logs of corresponding alphas
        """
        ############ Then we compute new points and densities ############
        std_normal = torch.distributions.Normal(loc=self.zero, scale=self.one)

        z_upd, update, eps, forward_grad = self._forward_step(z_old=z, x=x, target_logprob=target_logprob)

        target_log_density_upd = target_logprob(z=z_upd, x=x)
        target_log_density_old = target_logprob(z=z, x=x)

        eps_reverse = (-update - self.step_size * compute_grad(z=z_upd, target_logprob=target_logprob,
                                                               x=x)) / torch.sqrt(
            2 * self.step_size)
        proposal_density_numerator = std_normal.log_prob(eps_reverse).sum(1)
        proposal_density_denominator = std_normal.log_prob(eps).sum(1)

        log_t = target_log_density_upd - target_log_density_old - proposal_density_denominator + \
                proposal_density_numerator
        log_1pt = torch.logsumexp(torch.cat([torch.zeros_like(log_t).view(-1, 1),
                                             log_t.view(-1, 1)], dim=-1), dim=-1)  # log(1+t)

        a, current_log_alphas, _ = get_acceptance_ratio(log_t, log_1pt, use_barker=self.use_barker)

        z_new = torch.empty_like(z_upd)
        z_new[a] = z_upd[a]
        z_new[~a] = z[~a]

        return {"z_new": z_new,
                "decisions": a.to(torch.float32),
                "current_log_alphas": current_log_alphas,
                "forward_grad": forward_grad}


class ULA(BaseKernel):
    def __init__(self, transforms=None, ula_skip_threshold=0.0, **kwargs):
        """
        ULA kernel class

        :param step_size:
        :param learnable:
        :param transforms:
        :param ula_skip_threshold:
        """
        super().__init__(**kwargs)
        self.ula_skip_threshold = ula_skip_threshold
        self.name = "ULA"
        self.transforms = False
        self.add_nn = None
        self.scale_nn = None
        self.score_matching = False
        if transforms is not None:
            self.transforms = True
            self.add_nn = transforms()
            self.scale_nn = transforms()  ###just test with step size at the moment
            self.score_matching = True

    def _forward_step(self, z_old, x=None, target_logprob=None):
        eps = torch.randn_like(z_old)
        self.log_jac = torch.zeros_like(z_old[:, 0])
        if not self.transforms:
            add = torch.zeros_like(z_old)
            forward_grad = compute_grad(
                z=z_old,
                target_logprob=target_logprob,
                x=x)
            update = torch.sqrt(2 * self.step_size) * eps + self.step_size * forward_grad
            z_new = z_old + update
            eps_reverse = (z_old - z_new - self.step_size * compute_grad(z=z_new, target_logprob=target_logprob,
                                                                         x=x)) / torch.sqrt(
                2 * self.step_size)
            score_match_cur = add
        else:
            add = self.add_nn(z=z_old, x=x)
            z_new = z_old + self.step_size * add + torch.sqrt(2 * self.step_size) * eps
            eps_reverse = (z_old - z_new - self.step_size * self.add_nn(z=z_new, x=x)) / torch.sqrt(2 * self.step_size)
            score_match_cur = (add - compute_grad(z=z_old, target_logprob=target_logprob, x=x)) ** 2
            forward_grad = add
        return z_new, eps, eps_reverse, score_match_cur, forward_grad

    def scale_transform(self, z, sign='+'):
        S = torch.sigmoid(self.scale_nn(z))
        sign = {"+": 1., "-": -1.}[sign]
        self.log_jac += torch.sum(torch.log(S), dim=1) * sign
        return S

    def forward(self, z, target_logprob, x=None, reverse_kernel=None, mu_amortize=None, **kwargs):
        """
        Input:
        z_old - current position
        target_logprob - target_logprob distribution
        x - data object (optional)
        Output:
        z_new - new position
        current_log_alphas - current log_alphas, corresponding to sampled decision variables
        a - decision variables (0 or +1)
        """

        ############ Then we compute new points and densities ############
        std_normal = torch.distributions.Normal(loc=self.zero, scale=self.one)

        z_upd, eps, eps_reverse, score_match_cur, forward_grad = self._forward_step(z_old=z, x=x,
                                                                                    target_logprob=target_logprob)

        if reverse_kernel is None:
            proposal_density_numerator = std_normal.log_prob(eps_reverse).sum(1)
        else:
            mu, logvar = reverse_kernel(torch.cat([z_upd, mu_amortize], dim=1))
            proposal_density_numerator = torch.distributions.Normal(loc=mu + z_upd,
                                                                    scale=torch.exp(0.5 * logvar)).log_prob(
                z).sum(1)

        proposal_density_denominator = std_normal.log_prob(eps).sum(1)

        z_new = z_upd

        # ###
        # with torch.no_grad():
        #     target_log_density_upd = target_logprob(z=z_upd, x=x)
        #     target_log_density_old = target_logprob(z=z, x=x)
        #     log_t = target_log_density_upd + proposal_density_numerator - target_log_density_old - proposal_density_denominator + self.log_jac
        #     log_1pt = torch.logsumexp(torch.cat([torch.zeros_like(log_t).view(-1, 1),
        #                                          log_t.view(-1, 1)], dim=-1), dim=-1)  # log(1+t)
        #     if self.ula_skip_threshold > 0.:
        #         a, _, current_log_alphas_pre = get_acceptance_ratio(log_t, log_1pt, use_barker=False)
        #         acceptance_probs = torch.exp(current_log_alphas_pre)
        #         reject_mask = acceptance_probs <= self.ula_skip_threshold
        #     else:
        #         a, current_log_alphas, _ = get_acceptance_ratio(log_t, log_1pt, use_barker=False)
        #         reject_mask = torch.zeros_like(a) < -1.
        # ###
        # if reject_mask.sum():
        #     # z_new[reject_mask] = z[reject_mask]
        #     z_new = torch.where(reject_mask[..., None], z, z_new)
        #     proposal_density_numerator[reject_mask] = torch.zeros_like(proposal_density_numerator[reject_mask])
        #     proposal_density_denominator[reject_mask] = torch.zeros_like(proposal_density_denominator[reject_mask])
        #     score_match_cur[reject_mask] = torch.zeros_like(score_match_cur[reject_mask])

        return {"z_new": z_new,
                "aggregated_log_jac": self.log_jac,
                "aux_for_acceptance_ratio": proposal_density_numerator - proposal_density_denominator}



import torch
import torch.nn as nn

from metlibvi.mcmc import ULA
from metlibvi.trainers import train_met_vi, train_vi, metropolized_vi_forward
from metlibvi.vi import NormFlow

device = torch.device("cpu")  # "cuda" if torch.cuda.is_available() else "cpu")


def get_precision(samples, target):
    true_loc = target.loc
    sampled_loc = samples.mean(0)
    return torch.abs(true_loc - sampled_loc).mean()


def test_rnvp(get_gaussian_target):
    """
    Tests if RealNVP provides good VI
    :return:
    """
    target = get_gaussian_target
    nf = NormFlow(flow_type='RealNVP', num_flows=3, hidden_dim=2, need_permute=True, hidden_dims=[2]).to(device)
    prior = torch.distributions.MultivariateNormal(loc=torch.tensor([0., 0.], device=device),
                                                   covariance_matrix=torch.eye(2, device=device))

    nf = train_vi(transitions=nf, target=target, prior=prior, lr=7e-3, n_batches=5000)

    samples = nf(prior.sample((1000,)))["z_new"]
    inverse_samples = nf.inverse(target.sample((1000,)))["z_new"]
    assert get_precision(samples, target) <= 75e-2 and get_precision(inverse_samples, prior)


def test_met_ula(get_gaussian_target):
    """
    Tests is metropolized ULA provides good VI
    :param get_gaussian_target:
    :return:
    """
    target = get_gaussian_target
    transitions = nn.ModuleList([ULA(step_size=0.01, learnable=True) for _ in range(5)]).to(device)
    prior = torch.distributions.MultivariateNormal(loc=torch.tensor([0., 0.], device=device),
                                                   covariance_matrix=torch.eye(2, device=device))

    transitions, _ = train_met_vi(transitions=transitions, target=target, prior=prior, lr=7e-3)

    with torch.no_grad():
        samples = metropolized_vi_forward(transitions=transitions, target_logprob=target.log_prob,
                                          z=prior.sample((1000,)), )["z_new"]
    assert get_precision(samples, target) <= 75e-2


def test_met_realnvp(get_gaussian_target):
    """
    Tests is metropolized RealNVP (metflow) provides good VI
    :param get_gaussian_target:
    :return:
    """
    target = get_gaussian_target
    transitions = nn.ModuleList(
        [NormFlow(flow_type="RealNVP", num_flows=2, need_permute=True, hidden_dim=2, hidden_dims=(2,)) for _ in
         range(2)]).to(device)
    prior = torch.distributions.MultivariateNormal(loc=torch.tensor([0., 0.], device=device),
                                                   covariance_matrix=torch.eye(2, device=device))

    transitions, logit_p = train_met_vi(transitions=transitions, target=target, prior=prior, n_batches=10000, lr=7e-3)

    with torch.no_grad():
        samples = \
            metropolized_vi_forward(transitions=transitions, forward_logit_prob=logit_p, target_logprob=target.log_prob,
                                    z=prior.sample((1000,)), )["z_new"]

    assert get_precision(samples, target) <= 75e-2



import numpy as np
import torch
from PIL import Image
from torch.utils.data import Dataset, DataLoader
from torchvision import transforms, datasets


class MyDataset(Dataset):
    def __init__(self, data, binarize=False, reshape=False):
        super(MyDataset, self).__init__()
        if isinstance(data, list):
            if binarize:  # it is mnistlike datasets
                self.data = data[0]
                self.labels = data[1]
            else:
                self.data = data
                self.labels = None
        else:
            self.data = data
            self.labels = None
        self.binarize = binarize
        if isinstance(self.data, torch.Tensor) or isinstance(self.data, np.ndarray):
            self.shape_size = self.data.shape[-2]
        elif isinstance(self.data[0], str):
            self.shape_size = 64
        else:
            self.shape_size = self.data[0][0].size[-1]
        self.reshape = reshape

    def __len__(self):
        return len(self.data)

    def __getitem__(self, item):
        if isinstance(self.data[item], tuple):
            sample, _ = self.data[item]
        else:
            sample = self.data[item]
        if self.reshape:
            sample = Image.open(sample).convert("RGB")
            sample = transforms.Compose([
                transforms.Resize((self.shape_size, self.shape_size)),
                transforms.ToTensor(),
                # transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
            ])(sample)
        else:
            sample = transforms.ToTensor()(sample)
        if self.binarize:
            sample = torch.distributions.Bernoulli(probs=sample).sample()
        label = -1. if self.labels is None else self.labels[item]
        return sample, label


def make_dataloaders(dataset, batch_size, val_batch_size, binarize=False, **kwargs):
    if dataset == 'mnist':
        data = datasets.MNIST('./data', train=True, download=True)
        train_data = data.train_data.numpy()
        train_labels = data.train_labels.numpy()
        train_dataset = MyDataset([train_data, train_labels], binarize=binarize)
        train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True, **kwargs)

        val_data = data.test_data
        val_labels = data.test_labels
        val_dataset = MyDataset([val_data, val_labels], binarize=binarize)
        val_loader = DataLoader(val_dataset, batch_size=val_batch_size, shuffle=False, **kwargs)

    elif dataset == 'fashionmnist':
        train_data = datasets.FashionMNIST('./data', train=True, download=True).train_data.to(torch.float32).numpy()
        train_data /= train_data.max()
        train_dataset = MyDataset(train_data, binarize=binarize)
        train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True, **kwargs)

        val_data = datasets.FashionMNIST('./data', train=False).test_data.to(torch.float32).numpy()
        val_data /= val_data.max()
        val_dataset = MyDataset(val_data, binarize=binarize)
        val_loader = DataLoader(val_dataset, batch_size=val_batch_size, shuffle=False, **kwargs)

    elif dataset == 'cifar10':
        train_data = datasets.CIFAR10('./data', train=True, download=True)
        train_dataset = MyDataset(train_data, binarize=False)
        train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True, **kwargs)

        val_data = datasets.CIFAR10('./data', train=False, download=True)
        val_dataset = MyDataset(val_data, binarize=False)
        val_loader = DataLoader(val_dataset, batch_size=val_batch_size, shuffle=False, **kwargs)

    elif dataset == 'omniglot':
        train_data = datasets.Omniglot('./data', background=True, download=True)
        train_dataset = MyDataset(train_data, binarize=binarize)
        train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True, **kwargs)

        val_data = datasets.Omniglot('./data', background=False, download=True)
        val_dataset = MyDataset(val_data, binarize=binarize)
        val_loader = DataLoader(val_dataset, batch_size=val_batch_size, shuffle=False, **kwargs)

    else:
        raise ValueError

    return train_loader, val_loader



import matplotlib.pyplot as plt
import numpy as np
import torch


def plot_true_prop_accept(z_target=None, z_accepted=None, z_proposal=None, acceptance_logprobs=None, **kwargs):
    '''
    Plots scatter plot with target points, proposal and accepted (scaled according to acceptance ratio if provided)
    :param z_target: target samples
    :param z_proposal: proposal samples
    :param z_accepted: accepted samples
    :param acceptance_logprobs: acceptance logprobs
    :return: scatterplot
    '''
    plt.close()

    if acceptance_logprobs is not None:
        acceptance_probs = np.exp(acceptance_logprobs.cpu().detach().numpy())
    else:
        acceptance_probs = torch.ones_like(z_accepted[:, 1]).cpu().detach().numpy()

    plt.figure(**kwargs)

    if z_target is not None:
        z_target = z_target.cpu().detach().numpy()
        plt.scatter(x=z_target[:, 0], y=z_target[:, 1], c='b', label='Target')

    if z_proposal is not None:
        z_proposal = z_proposal.cpu().detach().numpy()
        plt.scatter(x=z_proposal[:, 0], y=z_proposal[:, 1], c='r', label='Proposal')

    if z_accepted is not None:
        z_accepted = z_accepted.cpu().detach().numpy()
        plt.scatter(x=z_accepted[:, 0], y=z_accepted[:, 1], c='g', label='Accepted', s=acceptance_probs + 15)

    plt.legend()
    plt.show()



import matplotlib.pyplot as plt
import numpy as np
import torch
from scipy.stats import kde


def plot_hist2d(samples, x_limits, y_limits, nbins=300, title=None, file_name=None, **kwargs):
    """
    The function plots 2d histogram for given samples
    :param samples: numpy array
    :param x_limits:
    :param y_limits:
    :param nbins:
    :param title:
    :param file_name: file name (if not provided, we dont save the picture)
    :param kwargs: arguments for plt.figure
    :return:
    """
    plt.close()
    plt.figure(**kwargs)

    x = samples[:, 0]
    y = samples[:, 1]
    k = kde.gaussian_kde([x, y], bw_method=0.1)
    xi, yi = np.mgrid[x_limits[0]:x_limits[1]:nbins * 1j, y_limits[0]:y_limits[1]:nbins * 1j]
    zi = k(np.vstack([xi.flatten(), yi.flatten()]))

    # Make the plot
    plt.pcolormesh(xi, yi, zi.reshape(xi.shape))
    plt.axis('off')
    plt.xlim((x_limits[0], x_limits[1]))
    plt.ylim((y_limits[0], y_limits[1]))
    if title is not None:
        plt.title(title)
    plt.tight_layout()
    if file_name is not None:
        plt.savefig(file_name, format='png')
    plt.show()


def plot_contours(target, x_limits, y_limits, npts=250, num_lines=6, device="cpu", **kwargs):
    plt.close()
    plt.figure(**kwargs)
    xside = np.linspace(-x_limits[0], x_limits[1], npts)
    yside = np.linspace(-y_limits[0], y_limits[1], npts)
    xx, yy = np.meshgrid(xside, yside)
    z = torch.tensor(np.hstack([xx.reshape(-1, 1), yy.reshape(-1, 1)]), dtype=torch.float32, device=device)
    log_probs = target.log_prob(z=z).cpu().detach().numpy()
    p = np.exp(log_probs).reshape(npts, npts)
    plt.contour(xx, yy, p, num_lines, colors='k')
    plt.axis('equal')



import torch

from metlibvi.mcmc import get_acceptance_ratio


def test_pre_acceptance_mh():
    """
    Tests if current_log_alphas_pre (acceptance prob if we would use it) is correct
    :return:
    """
    log_t = torch.tensor([torch.log(torch.tensor(0.8)), torch.log(torch.tensor(0.35))])
    log_1pt = torch.tensor([torch.log(torch.tensor(1.8)), torch.log(torch.tensor(1.35))])
    use_barker = False
    _, _, current_log_alphas_pre = get_acceptance_ratio(log_t.clone(), log_1pt.clone(), use_barker)
    assert sum(current_log_alphas_pre == torch.min(log_t, torch.zeros_like(log_t))) == 2


def test_pre_acceptance_barker():
    """
    Tests if current_log_alphas_pre (acceptance prob if we would use it) is correct
    :return:
    """
    log_t = torch.tensor([torch.log(torch.tensor(0.8)), torch.log(torch.tensor(0.35))])
    log_1pt = torch.tensor([torch.log(torch.tensor(1.8)), torch.log(torch.tensor(1.35))])
    use_barker = True
    _, _, current_log_alphas_pre = get_acceptance_ratio(log_t.clone(), log_1pt.clone(), use_barker)
    assert sum(current_log_alphas_pre == log_t - log_1pt) == 2



import torch

from metlibvi.mcmc import HMC, ULA, MALA, run_chain


def get_precision(kernel, target):
    true_loc = target.loc
    samples = run_chain(kernel=kernel, z_init=torch.zeros_like(true_loc)[None], target_logprob=target.log_prob,
                        n_steps=500,
                        burnin=500,
                        return_trace=True)
    sampled_loc = samples.mean(0)
    return torch.abs(true_loc - sampled_loc).mean()


# @pytest.mark.skip(reason='Not implemented yet')
def test_hmc_precision(get_gaussian_target):
    target = get_gaussian_target
    kernel = HMC(n_leapfrogs=3, step_size=0.1)
    assert get_precision(kernel, target) <= 75e-2


# @pytest.mark.skip(reason='Not implemented yet')
def test_mala_precision(get_gaussian_target):
    target = get_gaussian_target
    kernel = MALA(step_size=0.1)
    assert get_precision(kernel, target) <= 75e-2


# @pytest.mark.skip(reason='Not implemented yet')
def test_ula_precision(get_gaussian_target):
    target = get_gaussian_target
    kernel = ULA(step_size=0.1)
    assert get_precision(kernel, target) <= 75e-2
